#!/usr/bin/env python
#
#
#   Copyright (c) 2007 Jesse Guardiani <jesse@guardiani.us>
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License as
#   published by the Free Software Foundation; either version 2 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
#   02111-1307, USA.
#

import math, os, pygame, gc, time, sys
from pygame.locals import *
import gtk
import globals
from db      import db      as db
from theme   import theme   as theme


def which (filename):
  if not os.environ.has_key('PATH') or os.environ['PATH'] == '':
      p = os.defpath
  else:
      p = os.environ['PATH']
  pathlist = p.split (os.pathsep)
  for path in pathlist:
      f = os.path.join(path, filename)
      if os.access(f, os.X_OK):
          return f
  return None

class Manager():
  def __init__(self):
    ''' This is the brain of the application. It handles keeping track of all the widgets,
    knowing when to draw and when not to draw a given widget, feeding the widgets with event
    data, and much more. '''
    self.screen     = None
    self.background = None
    self.db         = None
    self.view       = None
    self.dirty      = False
    self.scrobbler  = None
    self.player     = None
    self.has_mplayer     = which('mplayer')
    self.has_a2dp        = which('a2dpd') and self.has_mplayer
    self.album_view      = None
    self.nowplaying_view = None
    self.volume_view     = None
    self.songposition_view = None
    self.play_button     = None
    self.screen_status   = "on" # is our display screen on, off, or dimmed?
    self.screen_manually_blanked  = False # have we manually blanked the screen? This is a forced state and tracked seperately.
    self.on_battery      = False # are we running on battery power or plugged in?
    self.has_focus       = True  # does our window have focus?
    self.idle_start      = 0
    self.idle_seconds    = 0
    self.first_time      = True
    self.enqueue_mode    = True # False = old (weird) functionality
    self.view_history    = []
    self.sleep_timer     = 0    # >0 = sleep timer enabled, time of last activity

  def __load_scrobbler(self):
    if globals.ISMAEMO:
      import maemoscrobbler
      self.scrobbler = maemoscrobbler.MaemoScrobbler()

  def __load_maemo(self):
    if globals.ISMAEMO:
      import widgets
      self.clear()
      status_area = widgets.Text("CONNECTING TO DBUS")
      status_area.rect.topleft = (130,200)
      status_area.update()
      status_area.draw(self.screen)
      pygame.display.flip()
      import maemo

  def _get_player(self):
    import widgets
    if globals.ISMAEMO:
      status_area = widgets.Text("INITIALIZING OSSO PLAYER")
      status_area.rect.topleft = (130,200)
      status_area.update()
      status_area.draw(self.screen)
      pygame.display.flip()
      import ossoplayer
      self.player = ossoplayer.OSSOPlayer(self.first_time)
    else:
      status_area = widgets.Text("INITIALIZING MPLAYER")
      status_area.rect.topleft = (130,200)
      status_area.update()
      status_area.draw(self.screen)
      pygame.display.flip()
      import mplayer
      self.player = mplayer.Mplayer()

    self.first_time=False

  #
  # Public Methods
  #
  def start(self,screen):
    import views,widgets,prefs,remote,playlist
    self.prefs    = prefs.Prefs()
    self.remote   = remote.Remote()
    self.playlist = playlist.Playlist()
    self.screen   = screen
    
    theme.set_theme(self.prefs.get('theme'))
    if manager.prefs.getBool("sleep_timer"): self.sleep_timer = 1

    globals.MAINFONTCOLOR   = theme.surface.get_at((93,512))
    globals.MAINFONTBGCOLOR = theme.surface.get_at((40,582))
    
    self.background = theme.get_image('background')

    self.screen.blit(self.background,(0,0))
    pygame.display.flip()
    self._get_player()
    
    self.__load_scrobbler()
    self.__load_maemo()
    self.menu_view         = views.MenuView()
    self.album_view        = views.AlbumView()
    self.nowplaying_button = widgets.NowPlayingButton()
    self.play_button       = widgets.PlayButton()
    self.a2dp_button       = widgets.A2DPButton()
    self.player_button     = widgets.PlayerButton()
    self.next_button       = widgets.NextButton()
    self.prev_button       = widgets.PrevButton()
    self.repeat_button     = widgets.RepeatButton()
    self.nowplaying_view   = views.NowPlayingView()
    self.volume_view       = views.VolumeView()
    self.songposition_view = views.SongPositionView()
    self.albumcover_view   = views.AlbumCoverView()
    self.show_menu()

  def handle_event(self,event):
    if not self.view.handle_event(event):
      return

    # global events
    self.nowplaying_button.handle_event(event)
    self.a2dp_button.handle_event(event)
    self.player_button.handle_event(event)
    self.play_button.handle_event(event)
    self.next_button.handle_event(event)
    self.prev_button.handle_event(event)
    self.repeat_button.handle_event(event)
    if event.type == KEYDOWN:
      if event.key == K_F7:    self.player.volume_increase()
      if event.key == K_F8:    self.player.volume_decrease()
      if event.key == K_LEFT:  self.prev_track()
      if event.key == K_RIGHT: self.next_track()
      if event.key == K_ESCAPE:self.back()
      if event.key == K_F4:    self.show_menu()

      # pause on fullscreen or headset button - headset button won't work yet (SDL's fault)
      # "X11: Unknown xsym, sym = 0x1008ff6e"
      if (event.key == K_F6) or (event.key == 0x1008ff6e): self.playlist.pause()
    if event.type == ACTIVEEVENT:
      self.has_focus = event.gain


  def view_history_pop(self):
    if not self.view_history:
      print "2 view_history=%s,view=%s" % (self.view_history,self.view)
      return None

    old_view = self.view
    new_view = self.view_history.pop()
    if old_view and old_view.is_temporary and not old_view in self.view_history: old_view.close()
    gc.collect()
    #print "here comes the garbage:\n\n"
    #print gc.garbage
    print "3 view_history=%s,view=%s" % (self.view_history,self.view)
    return new_view


  def back_by_one(self):
    new_view = self.view_history_pop()
    if not new_view: self.show_exit_confirm()
    else: self.view = new_view
    pygame.event.clear()


  def back(self,index=None):
    import widgets
    manager.clear()
    status_area = widgets.Text("LOADING")
    status_area.rect.topleft = (130,200)
    status_area.update()
    status_area.draw(manager.screen)
    pygame.display.flip()

    if index == None or index > len(self.view_history):
      self.back_by_one()
    else:
      while len(self.view_history) - 1 >= index:
        self.back_by_one()


  def update(self):
    if self.is_idle():
      if not self.idle_start: self.idle_start = time.time()
      self.idle_seconds=time.time()-self.idle_start
    else: self.idle_start = 0

    self.view.update()
    self.nowplaying_button.update()

  def draw(self):
    for widget in self.view.get_widgets(): widget.draw(self.screen)
    self.nowplaying_button.draw(self.screen)
    self.a2dp_button.draw(self.screen)
    self.player_button.draw(self.screen)
    self.play_button.draw(self.screen)
    self.next_button.draw(self.screen)
    self.prev_button.draw(self.screen)
    self.repeat_button.draw(self.screen)
    self.dirty = False

  def wake_up(self):
    manager.has_focus = True
    self.screen_manually_blanked = False
    self.idle_start   = 0
    self.idle_seconds = 0
    if self.sleep_timer>1:
      self.sleep_timer=1

  def is_idle(self):
    if self.screen_status == "off" or self.screen_manually_blanked: return True
    if self.dirty: return False
    for widget in self.view.get_widgets():
      if not widget.is_idle():
        return False
    return True

  def is_extended_idle(self):
    if globals.ISMAEMO:
      if self.screen_status == "off" or not self.has_focus or self.screen_manually_blanked:
        return True
    else:
      if self.idle_seconds > 5: return True
    return False

  def check_sleep_timer(self):
    if self.screen_status == "off" and self.player and not self.player.paused:
      if self.sleep_timer==1:
        self.sleep_timer = time.time()
      elif self.sleep_timer>1 and time.time() - self.sleep_timer >= 3600:
        print "sleep timer active!"
        self.player.pause(True)
        self.sleep_timer = 1    # in case of a remote unpause prevent immediate reactivation


  def update_powersave(self):
    gc.collect() # probably a good opportunity to run the GC as screen mode changes mess up the frame rate anyway.
    if self.screen_manually_blanked:
      if self.screen_status == "dimmed": globals.blank_screen()
      if self.screen_status == "on":
        print "screen on. disabling manual screen blank"
        self.screen_manually_blanked = False
    if self.player:
      if self.screen_status == "off" and self.on_battery: self.player.set_powersave(True)
      else: self.player.set_powersave(False)

  def quit(self):
    self.player.close()
    self.remote.close()
    self.playlist.save()
    self.prefs.save()

  def cleanup_views(self):
    if self.view and self.view.is_temporary and self.view: self.view.close()

  def show_menu(self):
    self.menu_view.show()

  def show_albums(self,selected_album=None):
    self.album_view.show()

  def show_artists(self):
    import views
    new_view = views.ArtistView()
    new_view.show()

  def show_playlists(self):
    import views
    new_view = views.PlaylistView()
    new_view.show()

  def show_m3u_songs(self,selected_m3u):
    import views
    if not selected_m3u: return
    print "showing songs for m3u=" + selected_m3u.name
    new_view = views.M3UView(selected_m3u.id,selected_m3u.name,selected_m3u.path)
    new_view.show()

  def show_all_songs(self):
    import views
    print "showing all songs"
    new_view = views.AllSongsView()
    new_view.show()

  def show_artist_songs(self,selected_artist):
    import views
    if not selected_artist: return
    print "showing songs for artist=" + selected_artist.artist.name
    new_view = views.ArtistAlbumView(selected_artist.artist_art.id,selected_artist.artist.name)
    new_view.show()

  def show_songs(self,selected_album):
    import views
    if not selected_album: return
    print "showing songs for album=" + selected_album.album.name
    new_view = views.SongView(selected_album.album.id,selected_album.album.name)
    new_view.show()

  def show_nowplaying(self,force=False):
    if not manager.playlist.list: return
    # This is a tri-mode button
    if force or self.view not in [self.nowplaying_view,self.albumcover_view]:
      print "showing nowplaying"
      self.nowplaying_view.show()
    elif self.view == self.nowplaying_view:
      self.show_albumcover_view()
    else:
      self.show_songposition_view()

  def show_albumcover_view(self,album_id=None):
    if self.playlist.current == None: return
    print "showing albumcover view"
    self.albumcover_view.show()
    sync_with_nowplaying = True if album_id == None else False
    self.view.update_image(album_id,sync_with_nowplaying=sync_with_nowplaying)

  def show_artistcover_view(self,album_id):
    print "showing artistcover view"
    import views
    new_view = views.ArtistCoverView()
    new_view.show()
    self.view.update_image(album_id)

  def show_songposition_view(self):
    if not self.songposition_view or not self.player.length: return # protect against calling before initialized
    print "showing song position view"
    self.songposition_view.show()
    self.view.update_selection()

  def show_volume(self):
    if not self.volume_view: return # protect against calling before initialized
    print "showing volume"
    if self.view == self.album_view or self.view == self.nowplaying_view: 
      self.view.scroll_widget.stop_animation()
    self.volume_view.show()
    self.view.update_selection()

  def show_exit_confirm(self):
    import views
    print "showing exit confirm"
    new_view = views.ExitConfirmView()
    new_view.show()

  def clear(self):
    self.screen.blit(self.background,(0,0))

  def prev_track(self):
    if self.player.seconds >= 5:
      print "prev: seeking to beginning"
      self.player.seek(0, 2)
      return
    self.playlist.prev(None,None)

  def next_track(self):
    if self.playlist.exhausted():
      print "no next track, ignored request"
      return
    self.playlist.next(None,None)
  

manager = Manager()
